#include "links.h"

#define WMAXBITS 15
#define ASCII_FLAG 0x01
#define HEAD_CRC 0x02
#define EXTRA_FIELD 0x04
#define ORIG_NAME 0x08
#define COMMENT 0x10
#define RESERVED 0xE0

static void *
z_mem_alloc(void *opaque, int items, int size)
{
   return mem_alloc(items * size);
}
   
static void z_mem_free(void *opaque, void *address)
{
   mem_free(address);
}
      


static z_streamp gzip_init(unsigned char *buf_old, int l)
{

/* check header */
/* gzip magic */
  unsigned char method;
  unsigned char flags;
  unsigned char *buf = buf_old;
  int len;
  int ret;
  z_streamp z;

  if (buf[0] != 0x1f || buf[1] != 0x8b) return NULL;

  method = buf[2];
  flags = buf[3];

  if (method != Z_DEFLATED || (flags & RESERVED) != 0) return NULL;

/* Comments are borrowed from gzio.c - zlib */
/* Discard time, xflags and OS code: */
  buf += 10;
  l -= 10;

  if ((flags & EXTRA_FIELD) != 0) { /* skip the extra field */
     len  =  2 + buf[0] + (buf[1] << 8);
     buf += len;
     l -= len;
  }
  if (l <= 0) return NULL;

  if ((flags & ORIG_NAME) != 0) {/* skip the original file name */
     len = strlen(buf) + 1;
     buf += len;
     l -= len;
  }
  if (l <= 0) return NULL;

  if ((flags & COMMENT) != 0) {/* skip the .gz file comment */
     len = strlen(buf) + 1;
     buf += len;
     l -= len;
  }
  if (l <= 0) return NULL;

  if ((flags & HEAD_CRC) != 0) {  /* skip the header crc */
     buf += 2;
     l -= 2;
  }
  if (l <= 0) return NULL;

/* initialization of z_stream */
  z = (z_streamp)mem_alloc(sizeof(z_stream));
  if (!z) return NULL;

  z->opaque = NULL;
  z->zalloc = (alloc_func)z_mem_alloc;
  z->zfree = (free_func)z_mem_free;
  z->avail_in = l;
  z->next_in = buf;
/* windowBits is passed < 0 to tell that there is no zlib header.
 * Note that in this case inflate *requires* an extra "dummy" byte
 * after the compressed stream in order to complete decompression and
 * return Z_STREAM_END. Here the gzip CRC32 ensures that 4 bytes are
 * present after the compressed stream.
 */
  ret = inflateInit2(z, -WMAXBITS);
  if (ret == Z_OK) return z;

  mem_free(z);
  return NULL;
}

#define OUTPUT_BUFFER_SIZE 65536

unsigned char *decompress_gzip(z_streamp *z, unsigned char *buf, int l, int *dl)
{
   unsigned char *output;
   int cur_size;
   int new_size;
   int ret;

   if (!*z) {
      *z = gzip_init(buf, l);
      if (!*z) {
         *dl = -1;
         return NULL;
      }
   }
   else {
      (*z)->next_in = buf;
      (*z)->avail_in = l;
   }
  
   (*z)->total_out = 0L;
   cur_size = OUTPUT_BUFFER_SIZE;
   output = mem_alloc(cur_size); /* output will be freed in http_read_data */
   if (!output) {
      inflateEnd(*z);
      mem_free(*z);
      *z = NULL;
      *dl = -2;
      return NULL;
   }

   (*z)->next_out = output;
   (*z)->avail_out = 65536;
   
   ret = inflate(*z, Z_SYNC_FLUSH);
   while (ret == Z_OK) {
      if (!(*z)->avail_in) {
         *dl = (int)(*z)->total_out;
         return output;
      }

      new_size = cur_size + OUTPUT_BUFFER_SIZE;
      output = mem_realloc(output, new_size);
      if (!output) {
         inflateEnd(*z);
         mem_free(*z);
         *z = NULL;
         *dl = -3;
         return NULL;
      }

      (*z)->next_out = output + cur_size; /* assume that z->avail_out == 0 */
      (*z)->avail_out = OUTPUT_BUFFER_SIZE;
      cur_size = new_size;
      ret = inflate(*z, Z_SYNC_FLUSH);
   }
   
   if (ret == Z_STREAM_END) *dl = (int)(*z)->total_out;
   else { /* something went wrong */
      *dl = -4;
      mem_free(output);
      output = NULL;
   }

   inflateEnd(*z);
   mem_free(*z);
   *z = NULL;
   return output;
}
